const char net_box_lua[] =
"-- net_box.lua (internal file)\n"
"local log      = require('log')\n"
"local ffi      = require('ffi')\n"
"local buffer   = require('buffer')\n"
"local socket   = require('socket')\n"
"local fiber    = require('fiber')\n"
"local msgpack  = require('msgpack')\n"
"local errno    = require('errno')\n"
"local urilib   = require('uri')\n"
"local internal = require('net.box.lib')\n"
"\n"
"local band          = bit.band\n"
"local max           = math.max\n"
"local fiber_time    = fiber.time\n"
"local fiber_self    = fiber.self\n"
"local ibuf_decode   = msgpack.ibuf_decode\n"
"\n"
"local table_new           = require('table.new')\n"
"local check_iterator_type = box.internal.check_iterator_type\n"
"\n"
"local communicate     = internal.communicate\n"
"local encode_auth     = internal.encode_auth\n"
"local encode_select   = internal.encode_select\n"
"local decode_greeting = internal.decode_greeting\n"
"\n"
"local sequence_mt      = { __serialize = 'sequence' }\n"
"local TIMEOUT_INFINITY = 500 * 365 * 86400\n"
"local VSPACE_ID        = 281\n"
"local VINDEX_ID        = 289\n"
"\n"
"local IPROTO_STATUS_KEY    = 0x00\n"
"local IPROTO_ERRNO_MASK    = 0x7FFF\n"
"local IPROTO_SYNC_KEY      = 0x01\n"
"local IPROTO_SCHEMA_ID_KEY = 0x05\n"
"local IPROTO_DATA_KEY      = 0x30\n"
"local IPROTO_ERROR_KEY     = 0x31\n"
"local IPROTO_GREETING_SIZE = 128\n"
"\n"
"-- select errors from box.error\n"
"local E_UNKNOWN              = box.error.UNKNOWN\n"
"local E_NO_CONNECTION        = box.error.NO_CONNECTION\n"
"local E_TIMEOUT              = box.error.TIMEOUT\n"
"local E_WRONG_SCHEMA_VERSION = box.error.WRONG_SCHEMA_VERSION\n"
"local E_PROC_LUA             = box.error.PROC_LUA\n"
"\n"
"-- utility tables\n"
"local is_final_state         = {closed = 1, error = 1}\n"
"local method_codec           = {\n"
"    ping    = internal.encode_ping,\n"
"    call_16 = internal.encode_call_16,\n"
"    call_17 = internal.encode_call,\n"
"    eval    = internal.encode_eval,\n"
"    insert  = internal.encode_insert,\n"
"    replace = internal.encode_replace,\n"
"    delete  = internal.encode_delete,\n"
"    update  = internal.encode_update,\n"
"    upsert  = internal.encode_upsert,\n"
"    select  = function(buf, id, schema_id, spaceno, indexno, key, opts)\n"
"        if type(spaceno) ~= 'number' then\n"
"            box.error(box.error.NO_SUCH_SPACE, '#'..tostring(spaceno))\n"
"        end\n"
"\n"
"        if type(indexno) ~= 'number' then\n"
"            box.error(box.error.NO_SUCH_INDEX, indexno, '#'..tostring(indexno))\n"
"        end\n"
"\n"
"        local limit = tonumber(opts and opts.limit) or 0xFFFFFFFF\n"
"        local offset = tonumber(opts and opts.offset) or 0\n"
"\n"
"        local key_is_nil = (key == nil or\n"
"                            (type(key) == 'table' and #key == 0))\n"
"        encode_select(buf, id, schema_id, spaceno, indexno,\n"
"                      check_iterator_type(opts, key_is_nil),\n"
"                      offset, limit, key)\n"
"    end,\n"
"    -- inject raw data into connection, used by console and tests\n"
"    inject = function(buf, id, schema_id, bytes)\n"
"        local ptr = buf:reserve(#bytes)\n"
"        ffi.copy(ptr, bytes, #bytes)\n"
"        buf.wpos = ptr + #bytes\n"
"    end\n"
"}\n"
"\n"
"local function next_id(id) return band(id + 1, 0x7FFFFFFF) end\n"
"\n"
"-- function create_transport(host, port, user, password, callback)\n"
"--\n"
"-- Transport methods: connect(), close(), perfrom_request(), wait_state()\n"
"--\n"
"-- Basically, *transport* is a TCP connection speaking one of\n"
"-- Tarantool network protocols. This is a low-level interface.\n"
"-- Primary features:\n"
"--  * implements protocols; concurrent perform_request()-s benefit from\n"
"--    multiplexing support in the protocol;\n"
"--  * schema-aware (optional) - snoops responses and initiates\n"
"--    schema reload when a request fails due to schema version mismatch;\n"
"--  * delivers transport events via the callback.\n"
"--\n"
"-- Transport state machine:\n"
"--\n"
"-- State machine starts in 'initial' state. Connect method changes\n"
"-- the state to 'connecting' and spawns a worker fiber. Close\n"
"-- method sets the state to 'closed' and kills the worker.\n"
"-- If the transport is already in 'error' state close() does nothing.\n"
"--\n"
"-- State chart:\n"
"--\n"
"--  initial -> connecting -> active\n"
"--                        \\\n"
"--                          -> auth -> fetch_schema <-> active\n"
"--\n"
"--  (any state, on error) -> error_reconnect -> connecting -> ...\n"
"--                                           \\\n"
"--                                             -> [error]\n"
"--  (any_state, but [error]) -> [closed]\n"
"--\n"
"--\n"
"-- State change events can be delivered to the transport user via\n"
"-- the optional 'callback' argument:\n"
"--\n"
"-- The callback functions needs to have the following signature:\n"
"--\n"
"--  callback(event_name, ...)\n"
"--\n"
"-- The following events are delivered, with arguments:\n"
"--\n"
"--  'state_changed', state, errno, error\n"
"--  'handshake', greeting           -> nil (accept) / errno, error (reject)\n"
"--  'will_fetch_schema'             -> true (approve) / false (skip fetch)\n"
"--  'did_fetch_schema', schema_id, spaces, indices\n"
"--  'will_reconnect', errno, error  -> true (approve) / false (reject)\n"
"--\n"
"-- Suggestion for callback writers: sleep a few secs before approving\n"
"-- reconnect.\n"
"--\n"
"local function create_transport(host, port, user, password, callback)\n"
"    -- check / normalize credentials\n"
"    if user == nil and password ~= nil then\n"
"        box.error(E_PROC_LUA, 'net.box: user is not defined')\n"
"    end\n"
"    if user ~= nil and password == nil then password = '' end\n"
"\n"
"    -- state: current state, only the worker fiber and connect method\n"
"    -- change state\n"
"    local state            = 'initial'\n"
"    local last_errno\n"
"    local last_error\n"
"    local state_cond       = fiber.cond() -- signaled when the state changes\n"
"\n"
"    -- requests: requests currently 'in flight', keyed by a request id;\n"
"    -- value refs are weak hence if a client dies unexpectedly,\n"
"    -- GC cleans the mess. Client submits a request and waits on state_cond.\n"
"    -- If the reponse arrives within the timeout, the worker wakes\n"
"    -- client fiber explicitly. Otherwize, wait on state_cond completes and\n"
"    -- the client reports E_TIMEOUT.\n"
"    local requests         = setmetatable({}, { __mode = 'v' })\n"
"    local next_request_id = 1\n"
"\n"
"    local worker_fiber\n"
"    local connection\n"
"    local send_buf         = buffer.ibuf(buffer.READAHEAD)\n"
"    local recv_buf         = buffer.ibuf(buffer.READAHEAD)\n"
"\n"
"    -- STATE SWITCHING --\n"
"    local function set_state(new_state, new_errno, new_error, schema_id)\n"
"        state = new_state\n"
"        last_errno = new_errno\n"
"        last_error = new_error\n"
"        callback('state_changed', new_state, new_errno, new_error)\n"
"        state_cond:broadcast()\n"
"        if state ~= 'active' then\n"
"            -- cancel all requests but the ones bearing the particular\n"
"            -- schema id; if schema id was omitted or we aren't fetching\n"
"            -- schema, cancel everything\n"
"            if not schema_id or state ~= 'fetch_schema' then\n"
"                schema_id = -1\n"
"            end\n"
"            local next_id, next_request = next(requests)\n"
"            while next_id do\n"
"                local id, request = next_id, next_request\n"
"                next_id, next_request = next(requests, id)\n"
"                if request.schema_id ~= schema_id then\n"
"                    requests[id] = nil -- this marks the request as completed\n"
"                    request.errno  = new_errno\n"
"                    request.response = new_error\n"
"                end\n"
"            end\n"
"        end\n"
"    end\n"
"\n"
"    -- FYI: [] on a string is valid\n"
"    local function wait_state(target_state, timeout)\n"
"        local deadline = fiber_time() + (timeout or TIMEOUT_INFINITY)\n"
"        repeat until state == target_state or target_state[state] or\n"
"                     is_final_state[state] or\n"
"                     not state_cond:wait(max(0, deadline - fiber_time()))\n"
"        return state == target_state or target_state[state] or false\n"
"    end\n"
"\n"
"    -- CONNECT/CLOSE --\n"
"    local protocol_sm\n"
"\n"
"    local function connect()\n"
"        if state ~= 'initial' then return not is_final_state[state] end\n"
"        set_state('connecting')\n"
"        fiber.create(function()\n"
"            worker_fiber = fiber_self()\n"
"            fiber.name(string.format('%s:%s (net.box)', host, port))\n"
"            local ok, err = pcall(protocol_sm)\n"
"            if not (ok or is_final_state[state]) then\n"
"                set_state('error', E_UNKNOWN, err)\n"
"            end\n"
"            if connection then\n"
"                connection:close()\n"
"                connection = nil\n"
"            end\n"
"            send_buf:recycle()\n"
"            recv_buf:recycle()\n"
"            worker_fiber = nil\n"
"        end)\n"
"        return true\n"
"    end\n"
"\n"
"    local function close()\n"
"        if not is_final_state[state] then\n"
"            set_state('closed', E_NO_CONNECTION, 'Connection closed')\n"
"        end\n"
"        if worker_fiber then\n"
"            worker_fiber:cancel()\n"
"            worker_fiber = nil\n"
"        end\n"
"    end\n"
"\n"
"    -- REQUEST/RESPONSE --\n"
"    local function perform_request(timeout, method, schema_id, ...)\n"
"        if state ~= 'active' then\n"
"            return last_errno or E_NO_CONNECTION, last_error\n"
"        end\n"
"        local deadline = fiber_time() + (timeout or TIMEOUT_INFINITY)\n"
"        -- alert worker to notify it of the queued outgoing data;\n"
"        -- if the buffer wasn't empty, assume the worker was already alerted\n"
"        if send_buf:size() == 0 then\n"
"            worker_fiber:wakeup()\n"
"        end\n"
"        local id = next_request_id\n"
"        method_codec[method](send_buf, id, schema_id, ...)\n"
"        next_request_id = next_id(id)\n"
"        local request = table_new(0, 5) -- reserve space for 5 keys\n"
"        request.client = fiber_self()\n"
"        request.method = method\n"
"        request.schema_id = schema_id\n"
"        requests[id] = request\n"
"        repeat\n"
"            local timeout = max(0, deadline - fiber_time())\n"
"            if not state_cond:wait(timeout) then\n"
"                requests[id] = nil\n"
"                return E_TIMEOUT, 'Timeout exceeded'\n"
"            end\n"
"        until requests[id] == nil -- i.e. completed (beware spurious wakeups)\n"
"        return request.errno, request.response\n"
"    end\n"
"\n"
"    local function dispatch_response(id, errno, response)\n"
"        local request = requests[id]\n"
"        if request then -- someone is waiting for the response\n"
"            requests[id] = nil\n"
"            request.errno, request.response = errno, response\n"
"            local client = request.client\n"
"            if client:status() ~= 'dead' then client:wakeup() end\n"
"        end\n"
"    end\n"
"\n"
"    local function dispatch_response_iproto(hdr, body)\n"
"        local id = hdr[IPROTO_SYNC_KEY]\n"
"        local status = hdr[IPROTO_STATUS_KEY]\n"
"        if status ~= 0 then\n"
"            return dispatch_response(id, band(status, IPROTO_ERRNO_MASK),\n"
"                                     body[IPROTO_ERROR_KEY])\n"
"        else\n"
"            return dispatch_response(id, nil, body[IPROTO_DATA_KEY])\n"
"        end\n"
"    end\n"
"\n"
"    local function new_request_id()\n"
"        local id = next_request_id;\n"
"        next_request_id = next_id(id)\n"
"        return id\n"
"    end\n"
"\n"
"    -- IO (WORKER FIBER) --\n"
"    local function send_and_recv(limit_or_boundary, timeout)\n"
"        return communicate(connection:fd(), send_buf, recv_buf,\n"
"                           limit_or_boundary, timeout)\n"
"    end\n"
"\n"
"    local function send_and_recv_iproto(timeout)\n"
"        local data_len = recv_buf.wpos - recv_buf.rpos\n"
"        local required = 0\n"
"        if data_len < 5 then\n"
"            required = 5\n"
"        else\n"
"            -- PWN! insufficient input validation\n"
"            local rpos, len = ibuf_decode(recv_buf.rpos)\n"
"            required = (rpos - recv_buf.rpos) + len\n"
"            if data_len >= required then\n"
"                local hdr\n"
"                rpos, hdr = ibuf_decode(rpos)\n"
"                local body = {}\n"
"                if rpos - recv_buf.rpos < required then\n"
"                    rpos, body = ibuf_decode(rpos)\n"
"                end\n"
"                recv_buf.rpos = rpos\n"
"                return nil, hdr, body\n"
"            end\n"
"        end\n"
"        local deadline = fiber_time() + (timeout or TIMEOUT_INFINITY)\n"
"        local err, extra = send_and_recv(required, timeout)\n"
"        if err then\n"
"            return err, extra\n"
"        end\n"
"        return send_and_recv_iproto(max(0, deadline - fiber_time()))\n"
"    end\n"
"\n"
"    local function send_and_recv_console(timeout)\n"
"        local delim = '\\n...\\n'\n"
"        local err, delim_pos = send_and_recv(delim, timeout)\n"
"        if err then\n"
"            return err, delim_pos\n"
"        else\n"
"            local response = ffi.string(recv_buf.rpos, delim_pos + #delim)\n"
"            recv_buf.rpos = recv_buf.rpos + delim_pos + #delim\n"
"            return nil, response\n"
"        end\n"
"    end\n"
"\n"
"    -- PROTOCOL STATE MACHINE (WORKER FIBER) --\n"
"    --\n"
"    -- The sm is implemented as a collection of functions performing\n"
"    -- tail-recursive calls to each other. Yep, Lua optimizes\n"
"    -- such calls, and yep, this is the canonical way to implement\n"
"    -- a state machine in Lua.\n"
"    local console_sm, iproto_auth_sm, iproto_schema_sm, iproto_sm, error_sm\n"
"\n"
"    protocol_sm = function ()\n"
"        connection = socket.tcp_connect(host, port)\n"
"        if connection == nil then\n"
"            return error_sm(E_NO_CONNECTION, errno.strerror(errno()))\n"
"        end\n"
"        local size = IPROTO_GREETING_SIZE\n"
"        local err, msg = send_and_recv(size, 0.3)\n"
"        if err then\n"
"            return error_sm(err, msg)\n"
"        end\n"
"        local g = decode_greeting(ffi.string(recv_buf.rpos, size))\n"
"        recv_buf.rpos = recv_buf.rpos + size\n"
"        if not g then\n"
"            return error_sm(E_NO_CONNECTION, 'Can\\'t decode handshake')\n"
"        end\n"
"        err, msg = callback('handshake', g)\n"
"        if err then\n"
"            return error_sm(err, msg)\n"
"        end\n"
"        if g.protocol == 'Lua console' then\n"
"            local setup_delimiter = 'require(\"console\").delimiter(\"$EOF$\")\\n'\n"
"            method_codec.inject(send_buf, nil, nil, setup_delimiter)\n"
"            local err, response = send_and_recv_console()\n"
"            if err then\n"
"                return error_sm(err, response)\n"
"            elseif response ~= '---\\n...\\n' then\n"
"                return error_sm(E_NO_CONNECTION, 'Unexpected response')\n"
"            end\n"
"            local rid = next_request_id\n"
"            set_state('active')\n"
"            return console_sm(rid)\n"
"        elseif g.protocol == 'Binary' then\n"
"            return iproto_auth_sm(g.salt)\n"
"        else\n"
"            return error_sm(E_NO_CONNECTION, 'Unknown protocol: ' .. g.protocol)\n"
"        end\n"
"    end\n"
"\n"
"    console_sm = function(rid)\n"
"        local delim = '\\n...\\n'\n"
"        local err, response = send_and_recv_console()\n"
"        if err then\n"
"            return error_sm(err, response)\n"
"        else\n"
"            dispatch_response(rid, nil, response)\n"
"            return console_sm(next_id(rid))\n"
"        end\n"
"    end\n"
"\n"
"    iproto_auth_sm = function(salt)\n"
"        set_state('auth')\n"
"        if not user or not password then\n"
"            set_state('fetch_schema')\n"
"            return iproto_schema_sm()\n"
"        end\n"
"        encode_auth(send_buf, new_request_id(), nil, user, password, salt)\n"
"        local err, hdr, body = send_and_recv_iproto()\n"
"        if err then\n"
"            return error_sm(err, hdr)\n"
"        end\n"
"        if hdr[IPROTO_STATUS_KEY] ~= 0 then\n"
"            return error_sm(E_NO_CONNECTION, body[IPROTO_ERROR_KEY])\n"
"        end\n"
"        set_state('fetch_schema')\n"
"        return iproto_schema_sm(hdr[IPROTO_SCHEMA_ID_KEY])\n"
"    end\n"
"\n"
"    iproto_schema_sm = function(schema_id)\n"
"        if not callback('will_fetch_schema') then\n"
"            set_state('active')\n"
"            return iproto_sm(schema_id)\n"
"        end\n"
"        local select1_id = new_request_id()\n"
"        local select2_id = new_request_id()\n"
"        local response = {}\n"
"        -- fetch everything from space _vspace, 2 = ITER_ALL\n"
"        encode_select(send_buf, select1_id, nil, VSPACE_ID, 0, 2, 0, 0xFFFFFFFF, nil)\n"
"        -- fetch everything from space _vindex, 2 = ITER_ALL\n"
"        encode_select(send_buf, select2_id, nil, VINDEX_ID, 0, 2, 0, 0xFFFFFFFF, nil)\n"
"        schema_id = nil -- any schema_id will do provided that\n"
"                        -- it is consistent across responses\n"
"        repeat\n"
"            local err, hdr, body = send_and_recv_iproto()\n"
"            if err then return error_sm(err, hdr) end\n"
"            dispatch_response_iproto(hdr, body)\n"
"            local id = hdr[IPROTO_SYNC_KEY]\n"
"            if id == select1_id or id == select2_id then\n"
"                -- response to a schema query we've submitted\n"
"                local status = hdr[IPROTO_STATUS_KEY]\n"
"                local response_schema_id = hdr[IPROTO_SCHEMA_ID_KEY]\n"
"                if status ~= 0 then\n"
"                    return error_sm(E_NO_CONNECTION, body[IPROTO_ERROR_KEY])\n"
"                end\n"
"                if schema_id == nil then\n"
"                    schema_id = response_schema_id\n"
"                elseif schema_id ~= response_schema_id then\n"
"                    -- schema changed while fetching schema; restart loader\n"
"                    return iproto_schema_sm()\n"
"                end\n"
"                response[id] = body[IPROTO_DATA_KEY]\n"
"            end\n"
"        until response[select1_id] and response[select2_id]\n"
"        callback('did_fetch_schema', schema_id,\n"
"                 response[select1_id], response[select2_id])\n"
"        set_state('active')\n"
"        return iproto_sm(schema_id)\n"
"    end\n"
"\n"
"    iproto_sm = function(schema_id)\n"
"        local err, hdr, body = send_and_recv_iproto()\n"
"        if err then return error_sm(err, hdr) end\n"
"        dispatch_response_iproto(hdr, body)\n"
"        local status = hdr[IPROTO_STATUS_KEY]\n"
"        local response_schema_id = hdr[IPROTO_SCHEMA_ID_KEY]\n"
"        if status ~= 0 and\n"
"           band(status, IPROTO_ERRNO_MASK) == E_WRONG_SCHEMA_VERSION and\n"
"           response_schema_id ~= schema_id\n"
"        then\n"
"            set_state('fetch_schema',\n"
"                      E_WRONG_SCHEMA_VERSION, body[IPROTO_ERROR_KEY],\n"
"                      response_schema_id)\n"
"            return iproto_schema_sm(schema_id)\n"
"        end\n"
"        return iproto_sm(schema_id)\n"
"    end\n"
"\n"
"    error_sm = function(err, msg)\n"
"        if connection then connection:close(); connection = nil end\n"
"        send_buf:recycle()\n"
"        recv_buf:recycle()\n"
"        set_state('error_reconnect', err, msg)\n"
"        if callback('will_reconnect', err, msg) then\n"
"            set_state('connecting')\n"
"            return protocol_sm()\n"
"        else\n"
"            set_state('error', err, msg)\n"
"        end\n"
"    end\n"
"\n"
"    return {\n"
"        close           = close,\n"
"        connect         = connect,\n"
"        wait_state      = wait_state,\n"
"        perform_request = perform_request\n"
"    }\n"
"end\n"
"\n"
"-- Wrap create_transport, adding auto-close-on-GC feature.\n"
"-- All the GC magic is neatly encapsulated!\n"
"-- The tricky part is the callback:\n"
"--  * callback (typically) references the transport (indirectly);\n"
"--  * worker fiber references the callback;\n"
"--  * fibers are GC roots - i.e. transport is never GC-ed!\n"
"-- We solve the issue by making the worker->callback ref weak.\n"
"-- Now it is necessary to have a strong ref to callback somewhere or\n"
"-- it is GC-ed prematurely. We wrap close() method, stashing the\n"
"-- ref in an upvalue (close() performance doesn't matter much.)\n"
"local create_transport = function(host, port, user, password, callback)\n"
"    local weak_refs = setmetatable({callback = callback}, {__mode = 'v'})\n"
"    local function weak_callback(...)\n"
"        local callback = weak_refs.callback\n"
"        if callback then return callback(...) end\n"
"    end\n"
"    local transport = create_transport(host, port, user,\n"
"                                       password, weak_callback)\n"
"    local transport_close = transport.close\n"
"    local gc_hook = ffi.gc(ffi.new('char[1]'), function()\n"
"        pcall(transport_close)\n"
"    end)\n"
"    transport.close = function()\n"
"        -- dummy gc_hook, callback refs prevent premature GC\n"
"        return transport_close(gc_hook, callback)\n"
"    end\n"
"    return transport\n"
"end\n"
"\n"
"local function parse_connect_params(host_or_uri, ...) -- self? host_or_uri port? opts?\n"
"    local port, opts = ...\n"
"    if type(host_or_uri) == 'table' then host_or_uri, port, opts = ... end\n"
"    if type(port) == 'table' then opts = port; port = nil end\n"
"    if opts == nil then opts = {} else\n"
"        local copy = {}\n"
"        for k, v in pairs(opts) do copy[k] = v end\n"
"        opts = copy\n"
"    end\n"
"    local host = host_or_uri\n"
"    if port == nil then\n"
"        local url = urilib.parse(tostring(host))\n"
"        if url == nil or url.service == nil then\n"
"            box.error(E_PROC_LUA,\n"
"                \"usage: connect(uri[, opts] | host, port[, opts])\")\n"
"        end\n"
"        host, port = url.host, url.service\n"
"        if opts.user == nil and opts.password == nil then\n"
"            opts.user, opts.password = url.login, url.password\n"
"        end\n"
"    end\n"
"    return host, port, opts\n"
"end\n"
"\n"
"local function remote_serialize(self)\n"
"    return {\n"
"        host = self.host,\n"
"        port = self.port,\n"
"        opts = next(self.opts) and self.opts,\n"
"        state = self.state,\n"
"        error = self.error,\n"
"        protocol = self.protocol,\n"
"        peer_uuid = self.peer_uuid,\n"
"        peer_version_id = self.peer_version_id\n"
"    }\n"
"end\n"
"\n"
"local remote_methods = {}\n"
"local remote_mt = {\n"
"    __index = remote_methods, __serialize = remote_serialize,\n"
"    __metatable = false\n"
"}\n"
"\n"
"local console_methods = {}\n"
"local console_mt = {\n"
"    __index = console_methods, __serialize = remote_serialize,\n"
"    __metatable = false\n"
"}\n"
"\n"
"local space_metatable, index_metatable\n"
"\n"
"local function connect(...)\n"
"    local host, port, opts = parse_connect_params(...)\n"
"    local user, password = opts.user, opts.password; opts.password = nil\n"
"    local remote = {host = host, port = port, opts = opts, state = 'initial'}\n"
"    local function callback(what, ...)\n"
"        if      what == 'state_changed' then\n"
"            local state, errno, err = ...\n"
"            remote.state, remote.error = state, err\n"
"            if state == 'error_reconnect' then\n"
"                log.warn('%s:%s: %s', host or \"\", port or \"\", err)\n"
"            end\n"
"        elseif what == 'handshake' then\n"
"            local greeting = ...\n"
"            if not opts.console and greeting.protocol ~= 'Binary' then\n"
"                return E_NO_CONNECTION,\n"
"                       'Unsupported protocol: '..greeting.protocol\n"
"            end\n"
"            remote.protocol = greeting.protocol\n"
"            remote.peer_uuid = greeting.uuid\n"
"            remote.peer_version_id = greeting.version_id\n"
"        elseif what == 'will_fetch_schema' then\n"
"            return not opts.console\n"
"        elseif what == 'did_fetch_schema' then\n"
"            remote:_install_schema(...)\n"
"        elseif what == 'will_reconnect' then\n"
"            if type(opts.reconnect_after) == 'number' then\n"
"                fiber.sleep(opts.reconnect_after); return true\n"
"            end\n"
"        end\n"
"    end\n"
"    if opts.console then\n"
"        setmetatable(remote, console_mt)\n"
"    else\n"
"        setmetatable(remote, remote_mt)\n"
"        remote._deadlines = setmetatable({}, {__mode = 'k'})\n"
"        remote._space_mt = space_metatable(remote)\n"
"        remote._index_mt = index_metatable(remote)\n"
"        if opts.call_16 then remote.call = remote.call_16 end\n"
"    end\n"
"    remote._transport = create_transport(host, port, user, password, callback)\n"
"    remote._transport.connect()\n"
"    if opts.wait_connected ~= false then\n"
"        remote._transport.wait_state('active', tonumber(opts.wait_connected))\n"
"    end\n"
"    return remote\n"
"end\n"
"\n"
"local function remote_check(remote, method)\n"
"    if type(remote) ~= 'table' then\n"
"        local fmt = 'Use remote:%s(...) instead of remote.%s(...):'\n"
"        box.error(E_PROC_LUA, string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"\n"
"function remote_methods:close()\n"
"    remote_check(self, 'close')\n"
"    self._transport.close()\n"
"end\n"
"\n"
"function remote_methods:is_connected()\n"
"    remote_check(self, 'is_connected')\n"
"    return self.state == 'active'\n"
"end\n"
"\n"
"function remote_methods:wait_connected(timeout)\n"
"    remote_check(self, 'wait_connected')\n"
"    return self._transport.wait_state('active', timeout)\n"
"end\n"
"\n"
"function remote_methods:_request(method, ...)\n"
"    local this_fiber = fiber_self()\n"
"    local transport = self._transport\n"
"    local perform_request = transport.perform_request\n"
"    local wait_state = transport.wait_state\n"
"    local deadlines = self._deadlines\n"
"    local deadline = deadlines[this_fiber]\n"
"    local err, res\n"
"    repeat\n"
"        local timeout = deadline and max(0, deadline - fiber_time())\n"
"        if self.state ~= 'active' then\n"
"            wait_state('active', timeout)\n"
"            timeout = deadline and max(0, deadline - fiber_time())\n"
"        end\n"
"        err, res = perform_request(timeout, method,\n"
"                                   self._schema_id, ...)\n"
"        if not err then\n"
"            setmetatable(res, sequence_mt)\n"
"            local postproc = method ~= 'eval' and method ~= 'call_17'\n"
"            if postproc and rawget(box, 'tuple') then\n"
"                local tnew = box.tuple.new\n"
"                for i, v in pairs(res) do\n"
"                    res[i] = tnew(v)\n"
"                end\n"
"            end\n"
"            return res\n"
"        elseif err == E_WRONG_SCHEMA_VERSION then\n"
"            err = nil\n"
"        end\n"
"    until err\n"
"    box.error({code = err, reason = res})\n"
"end\n"
"\n"
"function remote_methods:ping()\n"
"    remote_check(self, 'ping')\n"
"    local deadline = self._deadlines[fiber_self()]\n"
"    local timeout = deadline and max(0, deadline-fiber_time())\n"
"    local err = self._transport.perform_request(timeout, 'ping',\n"
"                                                self._schema_id)\n"
"    return not err or err == E_WRONG_SCHEMA_VERSION\n"
"end\n"
"\n"
"function remote_methods:reload_schema()\n"
"    remote_check(self, 'reload_schema')\n"
"    self:_request('select', VSPACE_ID, 0, nil, {limit = 0})\n"
"end\n"
"\n"
"function remote_methods:call_16(func_name, ...)\n"
"    remote_check(self, 'call')\n"
"    return self:_request('call_16', tostring(func_name), {...})\n"
"end\n"
"\n"
"function remote_methods:call(func_name, ...)\n"
"    remote_check(self, 'call')\n"
"    return unpack(self:_request('call_17', tostring(func_name), {...}))\n"
"end\n"
"\n"
"function remote_methods:eval(code, ...)\n"
"    remote_check(self, 'eval')\n"
"    return unpack(self:_request('eval', code, {...}))\n"
"end\n"
"\n"
"function remote_methods:wait_state(state, timeout)\n"
"    remote_check(self, 'wait_state')\n"
"    if timeout == nil then\n"
"        local deadline = self._deadlines[fiber_self()]\n"
"        timeout = deadline and max(0, deadline-fiber_time())\n"
"    end\n"
"    return self._transport.wait_state(state, timeout)\n"
"end\n"
"\n"
"function remote_methods:timeout(timeout)\n"
"    remote_check(self, 'timeout')\n"
"    self._deadlines[fiber_self()] = (timeout and fiber_time() + timeout)\n"
"    return self\n"
"end\n"
"\n"
"function remote_methods:_install_schema(schema_id, spaces, indices)\n"
"    local sl, space_mt, index_mt = {}, self._space_mt, self._index_mt\n"
"    for _, space in pairs(spaces) do\n"
"        local name = space[3]\n"
"        local id = space[1]\n"
"        local engine = space[4]\n"
"        local field_count = space[5]\n"
"\n"
"        local s = {\n"
"            id              = id,\n"
"            name            = name,\n"
"            engine          = engine,\n"
"            field_count     = field_count,\n"
"            enabled         = true,\n"
"            index           = {},\n"
"            temporary       = false\n"
"        }\n"
"        if #space > 5 then\n"
"            local opts = space[6]\n"
"            if type(opts) == 'table' then\n"
"                -- Tarantool >= 1.7.0\n"
"                s.temporary = not not opts.temporary\n"
"            elseif type(opts) == 'string' then\n"
"                -- Tarantool < 1.7.0\n"
"                s.temporary = string.match(opts, 'temporary') ~= nil\n"
"            end\n"
"        end\n"
"\n"
"        setmetatable(s, space_mt)\n"
"\n"
"        sl[id] = s\n"
"        sl[name] = s\n"
"    end\n"
"\n"
"    for _, index in pairs(indices) do\n"
"        local idx = {\n"
"            space   = index[1],\n"
"            id      = index[2],\n"
"            name    = index[3],\n"
"            type    = string.upper(index[4]),\n"
"            parts   = {},\n"
"        }\n"
"        local OPTS = 5\n"
"        local PARTS = 6\n"
"\n"
"        if type(index[OPTS]) == 'number' then\n"
"            idx.unique = index[OPTS] == 1\n"
"\n"
"            for k = 0, index[PARTS] - 1 do\n"
"                local pktype = index[7 + k * 2 + 1]\n"
"                local pkfield = index[7 + k * 2]\n"
"\n"
"                local pk = {\n"
"                    type = pktype,\n"
"                    fieldno = pkfield + 1\n"
"                }\n"
"                idx.parts[k + 1] = pk\n"
"            end\n"
"        else\n"
"            for k = 1, #index[PARTS] do\n"
"                local pktype = index[PARTS][k][2]\n"
"                local pkfield = index[PARTS][k][1]\n"
"\n"
"                local pk = {\n"
"                    type = pktype,\n"
"                    fieldno = pkfield + 1\n"
"                }\n"
"                idx.parts[k] = pk\n"
"            end\n"
"            idx.unique = not not index[OPTS].is_unique\n"
"        end\n"
"\n"
"        if sl[idx.space] ~= nil then\n"
"            sl[idx.space].index[idx.id] = idx\n"
"            sl[idx.space].index[idx.name] = idx\n"
"            idx.space = sl[idx.space]\n"
"            setmetatable(idx, index_mt)\n"
"        end\n"
"    end\n"
"\n"
"    self._schema_id = schema_id\n"
"    self.space = sl\n"
"end\n"
"\n"
"-- console methods\n"
"console_methods.close = remote_methods.close\n"
"console_methods.is_connected = remote_methods.is_connected\n"
"console_methods.wait_state = remote_methods.wait_state\n"
"function console_methods:eval(line, timeout)\n"
"    remote_check(self, 'eval')\n"
"    local err, res\n"
"    local transport = self._transport\n"
"    local pr = transport.perform_request\n"
"    if self.state ~= 'active' then\n"
"        local deadline = fiber_time() + (timeout or TIMEOUT_INFINITY)\n"
"        transport.wait_state('active', timeout)\n"
"        timeout = max(0, deadline - fiber_time())\n"
"    end\n"
"    if self.protocol == 'Binary' then\n"
"        local loader = 'return require(\"console\").eval(...)'\n"
"        err, res = pr(timeout, 'eval', nil, loader, {line})\n"
"    else\n"
"        assert(self.protocol == 'Lua console')\n"
"        err, res = pr(timeout, 'inject', nil, line..'$EOF$\\n')\n"
"    end\n"
"    if err then\n"
"        box.error({code = err, reason = res})\n"
"    end\n"
"    return res[1] or res\n"
"end\n"
"\n"
"-- space, index metatable\n"
"local function space_check(space, method)\n"
"    if type(space) ~= 'table' or space.id == nil then\n"
"        local fmt = 'Use space:%s(...) instead of space.%s(...)'\n"
"        box.error(E_PROC_LUA, string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"\n"
"local function one_tuple(tab)\n"
"    if tab[1] ~= nil then return tab[1] end\n"
"end\n"
"\n"
"space_metatable = function(remote)\n"
"    local methods = {}\n"
"\n"
"    function methods:insert(tuple)\n"
"        space_check(self, 'insert')\n"
"        return one_tuple(remote:_request('insert', self.id, tuple))\n"
"    end\n"
"\n"
"    function methods:replace(tuple)\n"
"        space_check(self, 'replace')\n"
"        return one_tuple(remote:_request('replace', self.id, tuple))\n"
"    end\n"
"\n"
"    function methods:select(key, opts)\n"
"        space_check(self, 'select')\n"
"        return remote:_request('select', self.id, 0, key, opts)\n"
"    end\n"
"\n"
"    function methods:delete(key)\n"
"        space_check(self, 'delete')\n"
"        return one_tuple(remote:_request('delete', self.id, 0, key))\n"
"    end\n"
"\n"
"    function methods:update(key, oplist)\n"
"        space_check(self, 'update')\n"
"        return one_tuple(remote:_request('update', self.id, 0, key, oplist))\n"
"    end\n"
"\n"
"    function methods:upsert(key, oplist)\n"
"        space_check(self, 'upsert')\n"
"        return one_tuple(remote:_request('upsert', self.id, 0, key, oplist))\n"
"    end\n"
"\n"
"    function methods:get(key)\n"
"        space_check(self, 'get')\n"
"        local res = remote:_request('select', self.id, 0, key,\n"
"                                    { limit = 2, iterator = 'EQ' })\n"
"        if res[2] ~= nil then box.error(box.error.MORE_THAN_ONE_TUPLE) end\n"
"        if res[1] ~= nil then return res[1] end\n"
"    end\n"
"\n"
"    return { __index = methods, __metatable = false }\n"
"end\n"
"\n"
"local function index_check(index, method)\n"
"    if type(index) ~= 'table' or index.id == nil then\n"
"        local fmt = 'Use index:%s(...) instead of index.%s(...)'\n"
"        box.error(E_PROC_LUA, string.format(fmt, method, method))\n"
"    end\n"
"end\n"
"\n"
"index_metatable = function(remote)\n"
"    local methods = {}\n"
"\n"
"    function methods:select(key, opts)\n"
"        index_check(self, 'select')\n"
"        return remote:_request('select', self.space.id, self.id, key, opts)\n"
"    end\n"
"\n"
"    function methods:get(key)\n"
"        index_check(self, 'get')\n"
"        local res = remote:_request('select', self.space.id, self.id, key,\n"
"                                    { limit = 2, iterator = 'EQ' })\n"
"        if res[2] ~= nil then box.error(box.error.MORE_THAN_ONE_TUPLE) end\n"
"        if res[1] ~= nil then return res[1] end\n"
"    end\n"
"\n"
"    function methods:min(key)\n"
"        index_check(self, 'min')\n"
"        local res = remote:_request('select', self.space.id, self.id, key,\n"
"                                    { limit = 1, iterator = 'GE' })\n"
"        return one_tuple(res)\n"
"    end\n"
"\n"
"    function methods:max(key)\n"
"        index_check(self, 'max')\n"
"        local res = remote:_request('select', self.space.id, self.id, key,\n"
"                                    { limit = 1, iterator = 'LE' })\n"
"        return one_tuple(res)\n"
"    end\n"
"\n"
"    function methods:count(key)\n"
"        index_check(self, 'count')\n"
"        local code = string.format('box.space.%s.index.%s:count',\n"
"                                   self.space.name, self.name)\n"
"        return remote:_request('call_16', code, { key })[1][1]\n"
"    end\n"
"\n"
"    function methods:delete(key)\n"
"        index_check(self, 'delete')\n"
"        local res = remote:_request('delete', self.space.id, self.id, key)\n"
"        return one_tuple(res)\n"
"    end\n"
"\n"
"    function methods:update(key, oplist)\n"
"        index_check(self, 'update')\n"
"        local res = remote:_request('update', self.space.id, self.id,\n"
"                                    key, oplist)\n"
"        return one_tuple(res)\n"
"    end\n"
"\n"
"    function methods:upsert(key, oplist)\n"
"        index_check(self, 'upsert')\n"
"        local res = remote:_request('upsert', self.space.id, self.id,\n"
"                                    key, oplist)\n"
"        return one_tuple(res)\n"
"    end\n"
"\n"
"    return { __index = methods, __metatable = false }\n"
"end\n"
"\n"
"local this_module = {\n"
"    create_transport = create_transport,\n"
"    connect = connect,\n"
"    new = connect -- Tarantool < 1.7.1 compatibility\n"
"}\n"
"\n"
"function this_module.timeout(timeout, ...)\n"
"    if type(timeout) == 'table' then timeout = ... end\n"
"    if not timeout then return this_module end\n"
"    local function timed_connect(...)\n"
"        local host, port, opts = parse_connect_params(...)\n"
"        if opts.wait_connected ~= false then opts.wait_connected = timeout end\n"
"        return connect(host, port, opts)\n"
"    end\n"
"    return setmetatable({\n"
"        connect = timed_connect, new = timed_connect\n"
"    }, {__index = this_module})\n"
"end\n"
"\n"
"local function rollback()\n"
"    if rawget(box, 'rollback') ~= nil then\n"
"        -- roll back local transaction on error\n"
"        box.rollback()\n"
"    end\n"
"end\n"
"\n"
"local function handle_eval_result(status, ...)\n"
"    if not status then\n"
"        rollback()\n"
"        return box.error(E_PROC_LUA, (...))\n"
"    end\n"
"    return ...\n"
"end\n"
"\n"
"this_module.self = {\n"
"    ping = function() return true end,\n"
"    reload_schema = function() end,\n"
"    close = function() end,\n"
"    timeout = function(self) return self end,\n"
"    wait_connected = function(self) return true end,\n"
"    is_connected = function(self) return true end,\n"
"    call = function(_box, proc_name, ...)\n"
"        if type(_box) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:call(proc_name, ...)\")\n"
"        end\n"
"        proc_name = tostring(proc_name)\n"
"        local status, proc, obj = pcall(package.loaded['box.internal'].\n"
"            call_loadproc, proc_name)\n"
"        if not status then\n"
"            rollback()\n"
"            return box.error() -- re-throw\n"
"        end\n"
"        local result\n"
"        if obj ~= nil then\n"
"            return handle_eval_result(pcall(proc, obj, ...))\n"
"        else\n"
"            return handle_eval_result(pcall(proc, ...))\n"
"        end\n"
"    end,\n"
"    eval = function(_box, expr, ...)\n"
"        if type(_box) ~= 'table' then\n"
"            box.error(box.error.PROC_LUA, \"usage: remote:eval(expr, ...)\")\n"
"        end\n"
"        local proc, errmsg = loadstring(expr)\n"
"        if not proc then\n"
"            proc, errmsg = loadstring(\"return \"..expr)\n"
"        end\n"
"        if not proc then\n"
"            rollback()\n"
"            return box.error(box.error.PROC_LUA, errmsg)\n"
"        end\n"
"        return handle_eval_result(pcall(proc, ...))\n"
"    end\n"
"}\n"
"\n"
"setmetatable(this_module.self, {\n"
"    __index = function(self, key)\n"
"        if key == 'space' then\n"
"            -- proxy self.space to box.space\n"
"            return require('box').space\n"
"        end\n"
"        return nil\n"
"    end\n"
"})\n"
"\n"
"package.loaded['net.box'] = this_module\n"
""
;
