const char socket_lua[] =
"-- socket.lua (internal file)\n"
"\n"
"local TIMEOUT_INFINITY      = 500 * 365 * 86400\n"
"local LIMIT_INFINITY = 2147483647\n"
"\n"
"local ffi = require('ffi')\n"
"local boxerrno = require('errno')\n"
"local internal = require('socket')\n"
"local fiber = require('fiber')\n"
"local fio = require('fio')\n"
"local log = require('log')\n"
"local buffer = require('buffer')\n"
"\n"
"local format = string.format\n"
"\n"
"ffi.cdef[[\n"
"    struct gc_socket {\n"
"        const int fd;\n"
"    };\n"
"    typedef uint32_t socklen_t;\n"
"    typedef ptrdiff_t ssize_t;\n"
"\n"
"    int connect(int sockfd, const struct sockaddr *addr,\n"
"                socklen_t addrlen);\n"
"    int bind(int sockfd, const struct sockaddr *addr,\n"
"                socklen_t addrlen);\n"
"\n"
"    ssize_t write(int fd, const char *octets, size_t len);\n"
"    ssize_t read(int fd, void *buf, size_t count);\n"
"    int listen(int fd, int backlog);\n"
"    int socket(int domain, int type, int protocol);\n"
"    int coio_close(int s);\n"
"    int shutdown(int s, int how);\n"
"    ssize_t send(int sockfd, const void *buf, size_t len, int flags);\n"
"    ssize_t recv(int s, void *buf, size_t len, int flags);\n"
"    int accept(int s, void *addr, void *addrlen);\n"
"    ssize_t sendto(int sockfd, const void *buf, size_t len, int flags,\n"
"                   const struct sockaddr *dest_addr, socklen_t addrlen);\n"
"\n"
"    int\n"
"    lbox_socket_local_resolve(const char *host, const char *port,\n"
"                         struct sockaddr *addr, socklen_t *socklen);\n"
"    int lbox_socket_nonblock(int fd, int mode);\n"
"\n"
"    int setsockopt(int s, int level, int iname, const void *opt, size_t optlen);\n"
"    int getsockopt(int s, int level, int iname, void *ptr, size_t *optlen);\n"
"\n"
"    typedef struct { int active; int timeout; } linger_t;\n"
"\n"
"    struct protoent {\n"
"        char  *p_name;       /* official protocol name */\n"
"        char **p_aliases;    /* alias list */\n"
"        int    p_proto;      /* protocol number */\n"
"    };\n"
"    struct protoent *getprotobyname(const char *name);\n"
"\n"
"    void *memmem(const void *haystack, size_t haystacklen,\n"
"        const void *needle, size_t needlelen);\n"
"]]\n"
"\n"
"local gc_socket_t = ffi.metatype(ffi.typeof('struct gc_socket'), {\n"
"    __gc = function (socket)\n"
"        if socket.fd < 0 then return end\n"
"        if ffi.C.coio_close(socket.fd) ~= 0 then\n"
"            log.error(\"socket: failed to close fd=%d on gc: %s\", socket.fd,\n"
"                boxerrno.strerror())\n"
"        end\n"
"    end\n"
"})\n"
"\n"
"local socket_mt\n"
"\n"
"local function check_socket(socket)\n"
"    local gc_socket = type(socket) == 'table' and socket._gc_socket\n"
"    if ffi.istype(gc_socket_t, gc_socket) then\n"
"        local fd = gc_socket.fd\n"
"        if fd >= 0 then\n"
"            return fd\n"
"        else\n"
"            error(\"attempt to use closed socket\")\n"
"        end\n"
"    else\n"
"        local msg = \"Usage: socket:method()\"\n"
"        if socket ~= nil then msg = msg .. \", called with non-socket\" end\n"
"        error(msg)\n"
"    end\n"
"end\n"
"\n"
"local function make_socket(fd)\n"
"    local socket = { _gc_socket = ffi.new(gc_socket_t, { fd = fd }) }\n"
"    return setmetatable(socket, socket_mt)\n"
"end\n"
"\n"
"local gc_socket_sentinel = ffi.new(gc_socket_t, { fd = -1 })\n"
"\n"
"local function close_socket(socket)\n"
"    local fd = check_socket(socket)\n"
"    socket._errno = nil\n"
"    local r = ffi.C.coio_close(fd)\n"
"    -- .fd is const to prevent tampering\n"
"    ffi.copy(socket._gc_socket, gc_socket_sentinel, ffi.sizeof(gc_socket_t))\n"
"    if r ~= 0 then\n"
"        socket._errno = boxerrno()\n"
"        return false\n"
"    end\n"
"    return true\n"
"end\n"
"\n"
"local function get_ivalue(table, key)\n"
"    if type(key) == 'number' then\n"
"        return key\n"
"    end\n"
"    return table[key]\n"
"end\n"
"\n"
"local function get_iflags(table, flags)\n"
"    if flags == nil then\n"
"        return 0\n"
"    end\n"
"    local res = 0\n"
"    if type(flags) ~= 'table' then\n"
"        flags = { flags }\n"
"    end\n"
"    for i, f in pairs(flags) do\n"
"        if table[f] == nil then\n"
"            return nil\n"
"        end\n"
"        res = bit.bor(res, table[f])\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"local socket_methods  = {}\n"
"socket_methods.errno = function(self)\n"
"    check_socket(self)\n"
"    if self['_errno'] == nil then\n"
"        return 0\n"
"    else\n"
"        return self['_errno']\n"
"    end\n"
"end\n"
"\n"
"socket_methods.error = function(self)\n"
"    check_socket(self)\n"
"    if self['_errno'] == nil then\n"
"        return nil\n"
"    else\n"
"        return boxerrno.strerror(self._errno)\n"
"    end\n"
"end\n"
"\n"
"-- addrbuf is equivalent to struct sockaddr_storage\n"
"local addrbuf = ffi.new('char[128]') -- enough to fit any address\n"
"local addr = ffi.cast('struct sockaddr *', addrbuf)\n"
"local addr_len = ffi.new('socklen_t[1]')\n"
"socket_methods.sysconnect = function(self, host, port)\n"
"    local fd = check_socket(self)\n"
"    self._errno = nil\n"
"\n"
"    host = tostring(host)\n"
"    port = tostring(port)\n"
"\n"
"    addr_len[0] = ffi.sizeof(addrbuf)\n"
"    local res = ffi.C.lbox_socket_local_resolve(host, port, addr, addr_len)\n"
"    if res == 0 then\n"
"        res = ffi.C.connect(fd, addr, addr_len[0]);\n"
"        if res == 0 then\n"
"            return true\n"
"        end\n"
"    end\n"
"    self._errno = boxerrno()\n"
"    return false\n"
"end\n"
"\n"
"local function syswrite(self, charptr, size)\n"
"    local fd = check_socket(self)\n"
"    self._errno = nil\n"
"    local done = ffi.C.write(fd, charptr, size)\n"
"    if done < 0 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"\n"
"    return tonumber(done)\n"
"end\n"
"\n"
"socket_methods.syswrite = function(self, arg1, arg2)\n"
"    -- TODO: ffi.istype('char *', arg1) doesn't work for ffi.new('char[256]')\n"
"    if type(arg1) == 'cdata' and arg2 ~= nil then\n"
"        return syswrite(self, arg1, arg2)\n"
"    elseif type(arg1) == 'string' then\n"
"        return syswrite(self, arg1, #arg1)\n"
"    else\n"
"        error('Usage: socket:syswrite(data) or socket:syswrite(const char *, size)')\n"
"    end\n"
"end\n"
"\n"
"local function sysread(self, charptr, size)\n"
"    local fd = check_socket(self)\n"
"\n"
"    self._errno = nil\n"
"    local res = ffi.C.read(fd, charptr, size)\n"
"    if res < 0 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"\n"
"    return tonumber(res)\n"
"end\n"
"\n"
"socket_methods.sysread = function(self, arg1, arg2)\n"
"    -- TODO: ffi.istype('char *', arg1) doesn't work for ffi.new('char[256]')\n"
"    if type(arg1) == 'cdata' and arg2 ~= nil then\n"
"        return sysread(self, arg1, arg2)\n"
"    end\n"
"\n"
"    local size = arg1 or buffer.READAHEAD\n"
"\n"
"    local buf = buffer.IBUF_SHARED\n"
"    buf:reset()\n"
"    local p = buf:alloc(size)\n"
"\n"
"    local res = sysread(self, p, size)\n"
"    if res then\n"
"        local str = ffi.string(p, res)\n"
"        buf:recycle()\n"
"        return str\n"
"    else\n"
"        buf:recycle()\n"
"        return res\n"
"    end\n"
"end\n"
"\n"
"socket_methods.nonblock = function(self, nb)\n"
"    local fd = check_socket(self)\n"
"    self._errno = nil\n"
"\n"
"    local res\n"
"\n"
"    if nb == nil then\n"
"        res = ffi.C.lbox_socket_nonblock(fd, 0x80)\n"
"    elseif nb then\n"
"        res = ffi.C.lbox_socket_nonblock(fd, 1)\n"
"    else\n"
"        res = ffi.C.lbox_socket_nonblock(fd, 0)\n"
"    end\n"
"\n"
"    if res < 0 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"\n"
"    if res == 1 then\n"
"        return true\n"
"    else\n"
"        return false\n"
"    end\n"
"end\n"
"\n"
"local function do_wait(self, what, timeout)\n"
"    local fd = check_socket(self)\n"
"\n"
"    self._errno = nil\n"
"    timeout = timeout or TIMEOUT_INFINITY\n"
"\n"
"    local res = internal.iowait(fd, what, timeout)\n"
"    if res == 0 then\n"
"        self._errno = boxerrno.ETIMEDOUT\n"
"        return 0\n"
"    end\n"
"    return res\n"
"end\n"
"\n"
"socket_methods.readable = function(self, timeout)\n"
"    return do_wait(self, 1, timeout) ~= 0\n"
"end\n"
"\n"
"socket_methods.writable = function(self, timeout)\n"
"    return do_wait(self, 2, timeout) ~= 0\n"
"end\n"
"\n"
"socket_methods.wait = function(self, timeout)\n"
"    return do_wait(self, 'RW', timeout)\n"
"end\n"
"\n"
"socket_methods.listen = function(self, backlog)\n"
"    local fd = check_socket(self)\n"
"    self._errno = nil\n"
"    if backlog == nil then\n"
"        backlog = 256\n"
"    end\n"
"    local res = ffi.C.listen(fd, backlog)\n"
"    if res < 0 then\n"
"        self._errno = boxerrno()\n"
"        return false\n"
"    end\n"
"    return true\n"
"end\n"
"\n"
"socket_methods.bind = function(self, host, port)\n"
"    local fd = check_socket(self)\n"
"    self._errno = nil\n"
"\n"
"    host = tostring(host)\n"
"    port = tostring(port)\n"
"\n"
"    addr_len[0] = ffi.sizeof(addrbuf)\n"
"    local res = ffi.C.lbox_socket_local_resolve(host, port, addr, addr_len)\n"
"    if res == 0 then\n"
"        res = ffi.C.bind(fd, addr, addr_len[0]);\n"
"    end\n"
"    if res == 0 then\n"
"        return true\n"
"    end\n"
"\n"
"    self._errno = boxerrno()\n"
"    return false\n"
"end\n"
"\n"
"socket_methods.close = close_socket\n"
"\n"
"socket_methods.shutdown = function(self, how)\n"
"    local fd = check_socket(self)\n"
"    local hvariants = {\n"
"        ['R']           = 0,\n"
"        ['READ']        = 0,\n"
"        ['W']           = 1,\n"
"        ['WRITE']       = 1,\n"
"        ['RW']          = 2,\n"
"        ['READ_WRITE']  = 2,\n"
"\n"
"        [0]             = 0,\n"
"        [1]             = 1,\n"
"        [2]             = 2\n"
"    }\n"
"    local ihow = hvariants[how]\n"
"\n"
"    if ihow == nil then\n"
"        ihow = 2\n"
"    end\n"
"    self._errno = nil\n"
"    if ffi.C.shutdown(fd, ihow) < 0 then\n"
"        self._errno = boxerrno()\n"
"        return false\n"
"    end\n"
"    return true\n"
"end\n"
"\n"
"socket_methods.setsockopt = function(self, level, name, value)\n"
"    local fd = check_socket(self)\n"
"\n"
"    local info = get_ivalue(internal.SO_OPT, name)\n"
"\n"
"    if info == nil then\n"
"        error(format(\"Unknown socket option name: %s\", tostring(name)))\n"
"    end\n"
"\n"
"    if not info.rw then\n"
"        error(format(\"Socket option %s is read only\", name))\n"
"    end\n"
"\n"
"    self._errno = nil\n"
"\n"
"    if type(level) == 'string' then\n"
"        if level == 'SOL_SOCKET' then\n"
"            level = internal.SOL_SOCKET\n"
"        else\n"
"            local p = ffi.C.getprotobyname(level)\n"
"            if p == nil then\n"
"                self._errno = boxerrno()\n"
"                return false\n"
"            end\n"
"            level = p.p_proto\n"
"        end\n"
"    else\n"
"        level = tonumber(level)\n"
"    end\n"
"\n"
"    if type(value) == 'boolean' then\n"
"        if value then\n"
"            value = 1\n"
"        else\n"
"            value = 0\n"
"        end\n"
"    end\n"
"\n"
"    if info.type == 1 then\n"
"        local value = ffi.new(\"int[1]\", value)\n"
"        local res = ffi.C.setsockopt(fd,\n"
"            level, info.iname, value, ffi.sizeof('int'))\n"
"\n"
"        if res < 0 then\n"
"            self._errno = boxerrno()\n"
"            return false\n"
"        end\n"
"        return true\n"
"    end\n"
"\n"
"    if info.type == 2 then\n"
"        local res = ffi.C.setsockopt(fd,\n"
"            level, info.iname, value, ffi.sizeof('size_t'))\n"
"        if res < 0 then\n"
"            self._errno = boxerrno()\n"
"            return false\n"
"        end\n"
"        return true\n"
"    end\n"
"\n"
"    if name == 'SO_LINGER' then\n"
"        error(\"Use s:linger(active[, timeout])\")\n"
"    end\n"
"    error(format(\"Unsupported socket option: %s\", name))\n"
"end\n"
"\n"
"socket_methods.getsockopt = function(self, level, name)\n"
"    local fd = check_socket(self)\n"
"\n"
"    local info = get_ivalue(internal.SO_OPT, name)\n"
"\n"
"    if info == nil then\n"
"        error(format(\"Unknown socket option name: %s\", tostring(name)))\n"
"    end\n"
"\n"
"    self._errno = nil\n"
"\n"
"    if type(level) == 'string' then\n"
"        if level == 'SOL_SOCKET' then\n"
"            level = internal.SOL_SOCKET\n"
"        else\n"
"            local p = ffi.C.getprotobyname(level)\n"
"            if p == nil then\n"
"                boxerrno(boxerrno.EINVAL)\n"
"                self._errno = boxerrno.EINVAL\n"
"                return nil\n"
"            end\n"
"            level = p.p_proto\n"
"        end\n"
"    else\n"
"        level = tonumber(level)\n"
"    end\n"
"\n"
"\n"
"    if info.type == 1 then\n"
"        local value = ffi.new(\"int[1]\", 0)\n"
"        local len = ffi.new(\"size_t[1]\", ffi.sizeof('int'))\n"
"        local res = ffi.C.getsockopt(fd, level, info.iname, value, len)\n"
"\n"
"        if res < 0 then\n"
"            self._errno = boxerrno()\n"
"            return nil\n"
"        end\n"
"\n"
"        if len[0] ~= 4 then\n"
"            error(format(\"Internal error: unexpected optlen: %d\", len[0]))\n"
"        end\n"
"        return tonumber(value[0])\n"
"    end\n"
"\n"
"    if info.type == 2 then\n"
"        local value = ffi.new(\"char[256]\", { 0 })\n"
"        local len = ffi.new(\"size_t[1]\", 256)\n"
"        local res = ffi.C.getsockopt(fd, level, info.iname, value, len)\n"
"        if res < 0 then\n"
"            self._errno = boxerrno()\n"
"            return nil\n"
"        end\n"
"        return ffi.string(value, tonumber(len[0]))\n"
"    end\n"
"\n"
"    if name == 'SO_LINGER' then\n"
"        error(\"Use s:linger()\")\n"
"    end\n"
"    error(format(\"Unsupported socket option: %s\", name))\n"
"end\n"
"\n"
"socket_methods.linger = function(self, active, timeout)\n"
"    local fd = check_socket(self)\n"
"\n"
"    local info = internal.SO_OPT.SO_LINGER\n"
"    self._errno = nil\n"
"    if active == nil then\n"
"        local value = ffi.new(\"linger_t[1]\")\n"
"        local len = ffi.new(\"size_t[1]\", 2 * ffi.sizeof('int'))\n"
"        local res = ffi.C.getsockopt(fd,\n"
"            internal.SOL_SOCKET, info.iname, value, len)\n"
"        if res < 0 then\n"
"            self._errno = boxerrno()\n"
"            return nil\n"
"        end\n"
"        if value[0].active ~= 0 then\n"
"            active = true\n"
"        else\n"
"            active = false\n"
"        end\n"
"        return active, value[0].timeout\n"
"    end\n"
"\n"
"    if timeout == nil then\n"
"        timeout = 0\n"
"    end\n"
"\n"
"    local iactive\n"
"    if active then\n"
"        iactive = 1\n"
"    else\n"
"        iactive = 0\n"
"    end\n"
"\n"
"    local value = ffi.new(\"linger_t[1]\",\n"
"        { { active = iactive, timeout = timeout } })\n"
"    local len = 2 * ffi.sizeof('int')\n"
"    local res = ffi.C.setsockopt(fd,\n"
"        internal.SOL_SOCKET, info.iname, value, len)\n"
"    if res < 0 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"\n"
"    return active, timeout\n"
"end\n"
"\n"
"socket_methods.accept = function(self)\n"
"    local server_fd = check_socket(self)\n"
"    self._errno = nil\n"
"\n"
"    local client_fd, from = internal.accept(server_fd)\n"
"    if client_fd == nil then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    local client = make_socket(client_fd)\n"
"    if not client:nonblock(true) then\n"
"        client:close()\n"
"        return\n"
"    end\n"
"    return client, from\n"
"end\n"
"\n"
"local errno_is_transient = {\n"
"    [boxerrno.EAGAIN] = true;\n"
"    [boxerrno.EWOULDBLOCK] = true;\n"
"    [boxerrno.EINTR] = true;\n"
"}\n"
"\n"
"local errno_is_fatal = {\n"
"    [boxerrno.EBADF] = true;\n"
"    [boxerrno.EINVAL] = true;\n"
"    [boxerrno.EOPNOTSUPP] = true;\n"
"    [boxerrno.ENOTSOCK] = true;\n"
"}\n"
"\n"
"local function check_limit(self, limit)\n"
"    if self.rbuf:size() >= limit then\n"
"        return limit\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"local function check_delimiter(self, limit, eols)\n"
"    if limit == 0 then\n"
"        return 0\n"
"    end\n"
"    local rbuf = self.rbuf\n"
"    if rbuf:size() == 0 then\n"
"        return nil\n"
"    end\n"
"\n"
"    local shortest\n"
"    for i, eol in ipairs(eols) do\n"
"        local data = ffi.C.memmem(rbuf.rpos, rbuf:size(), eol, #eol)\n"
"        if data ~= nil then\n"
"            local len = ffi.cast('char *', data) - rbuf.rpos + #eol\n"
"            if shortest == nil or shortest > len then\n"
"                shortest = len\n"
"            end\n"
"        end\n"
"    end\n"
"    if shortest ~= nil and shortest <= limit then\n"
"        return shortest\n"
"    elseif limit <= rbuf:size() then\n"
"        return limit\n"
"    end\n"
"    return nil\n"
"end\n"
"\n"
"local function read(self, limit, timeout, check, ...)\n"
"    assert(limit >= 0)\n"
"    limit = math.min(limit, LIMIT_INFINITY)\n"
"    local rbuf = self.rbuf\n"
"    if rbuf == nil then\n"
"        rbuf = buffer.ibuf()\n"
"        self.rbuf = rbuf\n"
"    end\n"
"\n"
"    local len = check(self, limit, ...)\n"
"    if len ~= nil then\n"
"        self._errno = nil\n"
"        local data = ffi.string(rbuf.rpos, len)\n"
"        rbuf.rpos = rbuf.rpos + len\n"
"        return data\n"
"    end\n"
"\n"
"    local started = fiber.time()\n"
"    while timeout > 0 do\n"
"        local started = fiber.time()\n"
"\n"
"        assert(rbuf:size() < limit)\n"
"        local to_read = math.min(limit - rbuf:size(), buffer.READAHEAD)\n"
"        local data = rbuf:reserve(to_read)\n"
"        assert(rbuf:unused() >= to_read)\n"
"        local res = sysread(self, data, rbuf:unused())\n"
"        if res == 0 then -- eof\n"
"            self._errno = nil\n"
"            local len = rbuf:size()\n"
"            local data = ffi.string(rbuf.rpos, len)\n"
"            rbuf.rpos = rbuf.rpos + len\n"
"            return data\n"
"        elseif res ~= nil then\n"
"            rbuf.wpos = rbuf.wpos + res\n"
"            local len = check(self, limit, ...)\n"
"            if len ~= nil then\n"
"                self._errno = nil\n"
"                local data = ffi.string(rbuf.rpos, len)\n"
"                rbuf.rpos = rbuf.rpos + len\n"
"                return data\n"
"            end\n"
"        elseif not errno_is_transient[self:errno()] then\n"
"            self._errno = boxerrno()\n"
"            return nil\n"
"        end\n"
"\n"
"        if not self:readable(timeout) then\n"
"            return nil\n"
"        end\n"
"        if timeout <= 0 then\n"
"            break\n"
"        end\n"
"        timeout = timeout - ( fiber.time() - started )\n"
"    end\n"
"    self._errno = boxerrno.ETIMEDOUT\n"
"    return nil\n"
"end\n"
"\n"
"socket_methods.read = function(self, opts, timeout)\n"
"    check_socket(self)\n"
"    timeout = timeout or TIMEOUT_INFINITY\n"
"    if type(opts) == 'number' then\n"
"        return read(self, opts, timeout, check_limit)\n"
"    elseif type(opts) == 'string' then\n"
"        return read(self, LIMIT_INFINITY, timeout, check_delimiter, { opts })\n"
"    elseif type(opts) == 'table' then\n"
"        local chunk = opts.chunk or opts.size or LIMIT_INFINITY\n"
"        local delimiter = opts.delimiter or opts.line\n"
"        if delimiter == nil then\n"
"            return read(self, chunk, timeout, check_limit)\n"
"        elseif type(delimiter) == 'string' then\n"
"            return read(self, chunk, timeout, check_delimiter, { delimiter })\n"
"        elseif type(delimiter) == 'table' then\n"
"            return read(self, chunk, timeout, check_delimiter, delimiter)\n"
"        end\n"
"    end\n"
"    error('Usage: s:read(delimiter|chunk|{delimiter = x, chunk = x}, timeout)')\n"
"end\n"
"\n"
"socket_methods.write = function(self, octets, timeout)\n"
"    check_socket(self)\n"
"    if timeout == nil then\n"
"        timeout = TIMEOUT_INFINITY\n"
"    end\n"
"\n"
"    local s = ffi.cast('const char *', octets)\n"
"    local p = s\n"
"    local e = s + #octets\n"
"    if p == e then\n"
"        return 0\n"
"    end\n"
"\n"
"    local started = fiber.time()\n"
"    while true do\n"
"        local written = syswrite(self, p, e - p)\n"
"        if written == 0 then\n"
"            return p - s -- eof\n"
"        elseif written ~= nil then\n"
"            p = p + written\n"
"            assert(p <= e)\n"
"            if p == e then\n"
"                return e - s\n"
"            end\n"
"        elseif not errno_is_transient[self:errno()] then\n"
"            return nil\n"
"        end\n"
"\n"
"        timeout = timeout - (fiber.time() - started)\n"
"        if timeout <= 0 or not self:writable(timeout) then\n"
"            break\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"socket_methods.send = function(self, octets, flags)\n"
"    local fd = check_socket(self)\n"
"    local iflags = get_iflags(internal.SEND_FLAGS, flags)\n"
"\n"
"    self._errno = nil\n"
"    local res = ffi.C.send(fd, octets, string.len(octets), iflags)\n"
"    if res < 0 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    return tonumber(res)\n"
"end\n"
"\n"
"socket_methods.recv = function(self, size, flags)\n"
"    local fd = check_socket(self)\n"
"    local iflags = get_iflags(internal.SEND_FLAGS, flags)\n"
"    if iflags == nil then\n"
"        self._errno = boxerrno.EINVAL\n"
"        return nil\n"
"    end\n"
"\n"
"    size = size or 512\n"
"    self._errno = nil\n"
"    local buf = ffi.new(\"char[?]\", size)\n"
"\n"
"    local res = ffi.C.recv(fd, buf, size, iflags)\n"
"\n"
"    if res == -1 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    return ffi.string(buf, res)\n"
"end\n"
"\n"
"socket_methods.recvfrom = function(self, size, flags)\n"
"    local fd = check_socket(self)\n"
"    local iflags = get_iflags(internal.SEND_FLAGS, flags)\n"
"    if iflags == nil then\n"
"        self._errno = boxerrno.EINVAL\n"
"        return nil\n"
"    end\n"
"\n"
"    size = size or 512\n"
"    self._errno = nil\n"
"    local res, from = internal.recvfrom(fd, size, iflags)\n"
"    if res == nil then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    return res, from\n"
"end\n"
"\n"
"socket_methods.sendto = function(self, host, port, octets, flags)\n"
"    local fd = check_socket(self)\n"
"    local iflags = get_iflags(internal.SEND_FLAGS, flags)\n"
"\n"
"    if iflags == nil then\n"
"        self._errno = boxerrno.EINVAL\n"
"        return nil\n"
"    end\n"
"\n"
"    self._errno = nil\n"
"    if octets == nil or octets == '' then\n"
"        return true\n"
"    end\n"
"\n"
"    host = tostring(host)\n"
"    port = tostring(port)\n"
"    octets = tostring(octets)\n"
"\n"
"    addr_len[0] = ffi.sizeof(addrbuf)\n"
"    local res = ffi.C.lbox_socket_local_resolve(host, port, addr, addr_len)\n"
"    if res == 0 then\n"
"        res = ffi.C.sendto(fd, octets, string.len(octets), iflags,\n"
"            addr, addr_len[0])\n"
"    end\n"
"    if res < 0 then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    return tonumber(res)\n"
"end\n"
"\n"
"local function create_socket(domain, stype, proto)\n"
"    local idomain = get_ivalue(internal.DOMAIN, domain)\n"
"    if idomain == nil then\n"
"        boxerrno(boxerrno.EINVAL)\n"
"        return nil\n"
"    end\n"
"\n"
"    local itype = get_ivalue(internal.SO_TYPE, stype)\n"
"    if itype == nil then\n"
"        boxerrno(boxerrno.EINVAL)\n"
"        return nil\n"
"    end\n"
"\n"
"    local iproto\n"
"    if type(proto) == 'string' then\n"
"        local p = ffi.C.getprotobyname(proto)\n"
"        if p == nil then\n"
"            boxerrno(boxerrno.EINVAL)\n"
"            return nil\n"
"        end\n"
"        iproto = p.p_proto\n"
"    else\n"
"        iproto = tonumber(proto)\n"
"    end\n"
"\n"
"    local fd = ffi.C.socket(idomain, itype, iproto)\n"
"    if fd >= 0 then\n"
"        local socket = make_socket(fd)\n"
"        if not socket:nonblock(true) then\n"
"            socket:close()\n"
"        else\n"
"            return socket\n"
"        end\n"
"    end\n"
"end\n"
"\n"
"local function getaddrinfo(host, port, timeout, opts)\n"
"    if type(timeout) == 'table' and opts == nil then\n"
"        opts = timeout\n"
"        timeout = TIMEOUT_INFINITY\n"
"    elseif timeout == nil then\n"
"        timeout = TIMEOUT_INFINITY\n"
"    end\n"
"    if port == nil then\n"
"        port = 0\n"
"    end\n"
"    local ga_opts = {}\n"
"    if opts ~= nil then\n"
"        if opts.type ~= nil then\n"
"            local itype = get_ivalue(internal.SO_TYPE, opts.type)\n"
"            if itype == nil then\n"
"                boxerrno(boxerrno.EINVAL)\n"
"                return nil\n"
"            end\n"
"            ga_opts.type = itype\n"
"        end\n"
"\n"
"        if opts.family ~= nil then\n"
"            local ifamily = get_ivalue(internal.DOMAIN, opts.family)\n"
"            if ifamily == nil then\n"
"                boxerrno(boxerrno.EINVAL)\n"
"                return nil\n"
"            end\n"
"            ga_opts.family = ifamily\n"
"        end\n"
"\n"
"        if opts.protocol ~= nil then\n"
"            local p = ffi.C.getprotobyname(opts.protocol)\n"
"            if p == nil then\n"
"                boxerrno(boxerrno.EINVAL)\n"
"                return nil\n"
"            end\n"
"            ga_opts.protocol = p.p_proto\n"
"        end\n"
"\n"
"        if opts.flags ~= nil then\n"
"            ga_opts.flags =\n"
"                get_iflags(internal.AI_FLAGS, opts.flags)\n"
"            if ga_opts.flags == nil then\n"
"                boxerrno(boxerrno.EINVAL)\n"
"                return nil\n"
"            end\n"
"        end\n"
"\n"
"    end\n"
"    return internal.getaddrinfo(host, port, timeout, ga_opts)\n"
"end\n"
"\n"
"local soname_mt = {\n"
"    __tostring = function(si)\n"
"        if si.host == nil and si.port == nil then\n"
"            return ''\n"
"        end\n"
"        if si.host == nil then\n"
"            return format('%s:%s', '0', tostring(si.port))\n"
"        end\n"
"\n"
"        if si.port == nil then\n"
"            return format('%s:%', tostring(si.host), 0)\n"
"        end\n"
"        return format('%s:%s', tostring(si.host), tostring(si.port))\n"
"    end\n"
"}\n"
"\n"
"socket_methods.name = function(self)\n"
"    local fd = check_socket(self)\n"
"    local aka = internal.name(fd)\n"
"    if aka == nil then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    self._errno = nil\n"
"    setmetatable(aka, soname_mt)\n"
"    return aka\n"
"end\n"
"\n"
"socket_methods.peer = function(self)\n"
"    local fd = check_socket(self)\n"
"    local peer = internal.peer(fd)\n"
"    if peer == nil then\n"
"        self._errno = boxerrno()\n"
"        return nil\n"
"    end\n"
"    self._errno = nil\n"
"    setmetatable(peer, soname_mt)\n"
"    return peer\n"
"end\n"
"\n"
"socket_methods.fd = function(self)\n"
"    return check_socket(self)\n"
"end\n"
"\n"
"-- tcp connector\n"
"local function tcp_connect_remote(remote, timeout)\n"
"    local s = create_socket(remote.family, remote.type, remote.protocol)\n"
"    if not s then\n"
"        -- Address family is not supported by the host\n"
"        return nil\n"
"    end\n"
"    local res = s:sysconnect(remote.host, remote.port)\n"
"    if res then\n"
"        -- Even through the socket is nonblocking, if the server to which we\n"
"        -- are connecting is on the same host, the connect is normally\n"
"        -- established immediately when we call connect (Stevens UNP).\n"
"        boxerrno(0)\n"
"        return s\n"
"    end\n"
"    local save_errno = s:errno()\n"
"    if save_errno ~= boxerrno.EINPROGRESS then\n"
"        s:close()\n"
"        boxerrno(save_errno)\n"
"        return nil\n"
"    end\n"
"    -- Wait until the connection is established or ultimately fails.\n"
"    -- In either condition the socket becomes writable. To tell these\n"
"    -- conditions appart SO_ERROR must be consulted (man connect).\n"
"    if s:writable(timeout) then\n"
"        save_errno = s:getsockopt('SOL_SOCKET', 'SO_ERROR')\n"
"    else\n"
"        save_errno = boxerrno.ETIMEDOUT\n"
"    end\n"
"    if save_errno ~= 0 then\n"
"        s:close()\n"
"        boxerrno(save_errno)\n"
"        return nil\n"
"    end\n"
"    -- Connected\n"
"    boxerrno(0)\n"
"    return s\n"
"end\n"
"\n"
"local function tcp_connect(host, port, timeout)\n"
"    if host == 'unix/' then\n"
"        return tcp_connect_remote({ host = host, port = port, protocol = 0,\n"
"            family = 'PF_UNIX', type = 'SOCK_STREAM' }, timeout)\n"
"    end\n"
"    local timeout = timeout or TIMEOUT_INFINITY\n"
"    local stop = fiber.time() + timeout\n"
"    local dns = getaddrinfo(host, port, timeout, { type = 'SOCK_STREAM',\n"
"        protocol = 'tcp' })\n"
"    if dns == nil or #dns == 0 then\n"
"        boxerrno(boxerrno.EINVAL)\n"
"        return nil\n"
"    end\n"
"    for i, remote in pairs(dns) do\n"
"        timeout = stop - fiber.time()\n"
"        if timeout <= 0 then\n"
"            boxerrno(boxerrno.ETIMEDOUT)\n"
"            return nil\n"
"        end\n"
"        local s = tcp_connect_remote(remote, timeout)\n"
"        if s then\n"
"            return s\n"
"        end\n"
"    end\n"
"    -- errno is set by tcp_connect_remote()\n"
"    return nil\n"
"end\n"
"\n"
"local function tcp_server_handler(server, sc, from)\n"
"    fiber.name(format(\"%s/%s:%s\", server.name, from.host, from.port))\n"
"    server.handler(sc, from)\n"
"    sc:shutdown()\n"
"    sc:close()\n"
"end\n"
"\n"
"local function tcp_server_loop(server, s, addr)\n"
"    fiber.name(format(\"%s/%s:%s\", server.name, addr.host, addr.port))\n"
"    log.info(\"started\")\n"
"    while s:readable() do\n"
"        local sc, from = s:accept()\n"
"        if sc == nil then\n"
"            local errno = s:errno()\n"
"            if not errno_is_transient[errno] then\n"
"                log.error('accept(%s) failed: %s', tostring(s), s:error())\n"
"            end\n"
"            if  errno_is_fatal[errno] then\n"
"                break\n"
"            end\n"
"        else\n"
"            fiber.create(tcp_server_handler, server, sc, from)\n"
"        end\n"
"    end\n"
"    -- Socket was closed\n"
"    if addr.family == 'AF_UNIX' and addr.port then\n"
"        fio.unlink(addr.port) -- remove unix socket\n"
"    end\n"
"    log.info(\"stopped\")\n"
"end\n"
"\n"
"local function tcp_server_usage()\n"
"    error('Usage: socket.tcp_server(host, port, handler | opts)')\n"
"end\n"
"\n"
"local function tcp_server_bind(s, addr)\n"
"    if s:bind(addr.host, addr.port) then\n"
"        return true\n"
"    end\n"
"\n"
"    if addr.family ~= 'AF_UNIX' then\n"
"        return false\n"
"    end\n"
"\n"
"    if boxerrno() ~= boxerrno.EADDRINUSE then\n"
"        return false\n"
"    end\n"
"\n"
"    local save_errno = boxerrno()\n"
"\n"
"    local sc = tcp_connect(addr.host, addr.port)\n"
"    if sc ~= nil then\n"
"        sc:close()\n"
"        boxerrno(save_errno)\n"
"        return false\n"
"    end\n"
"\n"
"    if boxerrno() ~= boxerrno.ECONNREFUSED then\n"
"        boxerrno(save_errno)\n"
"        return false\n"
"    end\n"
"\n"
"    log.info(\"tcp_server: remove dead UNIX socket: %s\", addr.port)\n"
"    if not fio.unlink(addr.port) then\n"
"        log.warn(\"tcp_server: %s\", boxerrno.strerror())\n"
"        boxerrno(save_errno)\n"
"        return false\n"
"    end\n"
"    return s:bind(addr.host, addr.port)\n"
"end\n"
"\n"
"\n"
"local function tcp_server(host, port, opts, timeout)\n"
"    local server = {}\n"
"    if type(opts) == 'function' then\n"
"        server.handler = opts\n"
"    elseif type(opts) == 'table' then\n"
"        if type(opts.handler) ~='function' or (opts.prepare ~= nil and\n"
"            type(opts.prepare) ~= 'function') then\n"
"            tcp_server_usage()\n"
"        end\n"
"        for k, v in pairs(opts) do\n"
"            server[k] = v\n"
"        end\n"
"    else\n"
"        tcp_server_usage()\n"
"    end\n"
"    server.name = server.name or 'server'\n"
"    timeout = timeout and tonumber(timeout) or TIMEOUT_INFINITY\n"
"    local dns\n"
"    if host == 'unix/' then\n"
"        dns = {{host = host, port = port, family = 'AF_UNIX', protocol = 0,\n"
"            type = 'SOCK_STREAM' }}\n"
"    else\n"
"        dns = getaddrinfo(host, port, timeout, { type = 'SOCK_STREAM',\n"
"            flags = 'AI_PASSIVE'})\n"
"        if dns == nil then\n"
"            return nil\n"
"        end\n"
"    end\n"
"\n"
"    for _, addr in ipairs(dns) do\n"
"        local s = create_socket(addr.family, addr.type, addr.protocol)\n"
"        if s ~= nil then\n"
"            local backlog\n"
"            if server.prepare then\n"
"                backlog = server.prepare(s)\n"
"            else\n"
"                s:setsockopt('SOL_SOCKET', 'SO_REUSEADDR', 1) -- ignore error\n"
"            end\n"
"            if not tcp_server_bind(s, addr) or not s:listen(backlog) then\n"
"                local save_errno = boxerrno()\n"
"                s:close()\n"
"                boxerrno(save_errno)\n"
"                return nil\n"
"            end\n"
"            fiber.create(tcp_server_loop, server, s, addr)\n"
"            return s, addr\n"
"       end\n"
"    end\n"
"    -- DNS resolved successfully, but addresss family is not supported\n"
"    boxerrno(boxerrno.EAFNOSUPPORT)\n"
"    return nil\n"
"end\n"
"\n"
"socket_mt   = {\n"
"    __index     = socket_methods,\n"
"    __tostring  = function(self)\n"
"        local fd = check_socket(self)\n"
"\n"
"        local save_errno = self._errno\n"
"        local name = format(\"fd %d\", fd)\n"
"        local aka = self:name()\n"
"        if aka ~= nil then\n"
"            name = format(\"%s, aka %s:%s\", name, aka.host, aka.port)\n"
"        end\n"
"        local peer = self:peer(self)\n"
"        if peer ~= nil then\n"
"            name = format(\"%s, peer %s:%s\", name, peer.host, peer.port)\n"
"        end\n"
"        self._errno = save_errno\n"
"        return name\n"
"    end,\n"
"    __serialize = function(self)\n"
"        -- Allow YAML, MsgPack and JSON to dump objects with sockets\n"
"        local fd = check_socket(self)\n"
"        return { fd = fd, peer = self:peer(), name = self:name() }\n"
"    end\n"
"}\n"
"\n"
"return setmetatable({\n"
"    getaddrinfo = getaddrinfo,\n"
"    tcp_connect = tcp_connect,\n"
"    tcp_server = tcp_server,\n"
"    iowait = internal.iowait\n"
"}, {\n"
"    __call = function(self, ...) return create_socket(...) end;\n"
"})\n"
""
;
